/*
 * FADE - Secure Overlay Cloud Storage with File Assured Deletion
 * Copyright (C) 2010  Yang Tang
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file Client.cc
 * @author Yang Tang
 * @brief FADE client implementation
 */

#include <cerrno>
#include <cstring>
#include <cstdio>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <openssl/aes.h>
#include <openssl/rand.h>
#include <openssl/hmac.h>
#include <openssl/sha.h>
#include <sys/stat.h>
#include "Client.h"
#include "PolicyForClient.h"
#include "../common/Panic.h"
#include "../common/Common.h"

static const int FILE_BUFFER_LENGTH = AES_BLOCK_SIZE << 8;

/**
 * @brief A chain of conjunctive keys
 */
struct Keychain
{
	unsigned char *key;
	Keychain *next;
};

Client *Client::instance()
{
	static Client _instance;
	return &_instance;
}

unsigned int Client::interact(const unsigned char *request, unsigned int requestLength, unsigned char *response, unsigned int responseLength)
{
	sockaddr_in addr;
	addr.sin_port = htons(DEFAULT_PORT);
	addr.sin_family = AF_INET;
	inet_pton(AF_INET, ephemerizerAddress.c_str(), &addr.sin_addr);
	if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		Panic::instance()->die(strerror(errno));
	}
	if (connect(sockfd, reinterpret_cast<sockaddr *>(&addr), sizeof(addr)))
	{
		closeSocket();
		Panic::instance()->die(strerror(errno));
	}
	if (send(sockfd, request, requestLength, 0) < 0)
	{
		Panic::instance()->warn(strerror(errno));
	}
	unsigned int len = recv(sockfd, response, responseLength, 0);
	closeSocket();
	return len;
}

void Client::closeSocket() const
{
	if (close(sockfd))
	{
		Panic::instance()->die(strerror(errno));
	}
}

void Client::readSecret()
{
	FILE *fp = fopen(secretFile.c_str(), "r");
	if (!fp)
	{
		Panic::instance()->warn("Cannot open secret file. Use GENSECRET command to generate one.\n");
		return;
	}
	fread(secret, sizeof(unsigned char), SECRET_LENGTH, fp);
	fclose(fp);
}

void Client::encrypt(const string &filename, const unsigned char *key) const
{
	AES_KEY aes;
	unsigned char iv[AES_BLOCK_SIZE];
	const char *pathname = (path + '/' + filename).c_str();
	FILE *ifp = fopen(pathname, "r");
	FILE *ofp = fopen((path + '/' + filename + ".fade").c_str(), "w");
	struct stat st;
	stat(pathname, &st);
	fwrite(&st.st_size, sizeof(st.st_size), 1, ofp);
	unsigned char iBuffer[FILE_BUFFER_LENGTH];
	unsigned char oBuffer[FILE_BUFFER_LENGTH];
	unsigned int len;
	HMAC_CTX ctx;
	HMAC_CTX_init(&ctx);
	HMAC_Init_ex(&ctx, secret, SECRET_LENGTH, EVP_sha1(), NULL);
	memset(iv, 0, sizeof(iv));
	if (AES_set_encrypt_key(key, 128, &aes) < 0)
	{
		Panic::instance()->die("Unable to set encryption key in AES.\n");
	}
	while ((len = fread(iBuffer, sizeof(unsigned char), FILE_BUFFER_LENGTH, ifp)) > 0)
	{
		if (len % AES_BLOCK_SIZE != 0)
		{
			len = (len / AES_BLOCK_SIZE + 1) * AES_BLOCK_SIZE;
		}
		AES_cbc_encrypt(iBuffer, oBuffer, len, &aes, iv, AES_ENCRYPT);
		fwrite(oBuffer, sizeof(unsigned char), len, ofp);
		HMAC_Update(&ctx, oBuffer, len);
	}
	fclose(ifp);
	unsigned char md[SHA_DIGEST_LENGTH];
	HMAC_Final(&ctx, md, &len);
	HMAC_CTX_cleanup(&ctx);
	fwrite(md, sizeof(unsigned char), len, ofp);
	fclose(ofp);
}

void Client::decrypt(const string &filename, const unsigned char *key) const
{
	AES_KEY aes;
	unsigned char iv[AES_BLOCK_SIZE];
	FILE *ifp = fopen((path + '/' + filename + ".fade").c_str(), "r");
	off_t fileSize;
	fread(&fileSize, sizeof(fileSize), 1, ifp);
	off_t fileSizeB = (fileSize % AES_BLOCK_SIZE == 0) ? fileSize : ((fileSize / AES_BLOCK_SIZE + 1) * AES_BLOCK_SIZE);
	FILE *ofp = fopen((path + '/' + filename).c_str(), "w");
	unsigned char iBuffer[FILE_BUFFER_LENGTH];
	unsigned char oBuffer[FILE_BUFFER_LENGTH];
	unsigned int len;
	memset(iv, 0, sizeof(iv));
	if (AES_set_decrypt_key(key, 128, &aes) < 0)
	{
		Panic::instance()->die("Unable to set decryption key in AES.\n");
	}
	while ((len = fread(iBuffer, sizeof(unsigned char), FILE_BUFFER_LENGTH < fileSizeB ? FILE_BUFFER_LENGTH : fileSizeB, ifp)) > 0)
	{
		AES_cbc_encrypt(iBuffer, oBuffer, len, &aes, iv, AES_DECRYPT);
		fwrite(oBuffer, sizeof(unsigned char), len < fileSize ? len : fileSize, ofp);
		fileSize -= len;
		fileSizeB -= len;
	}
	fclose(ifp);
	fclose(ofp);
}

bool Client::checkHMAC(const string &filename) const
{
	HMAC_CTX ctx;
	HMAC_CTX_init(&ctx);
	HMAC_Init_ex(&ctx, secret, SECRET_LENGTH, EVP_sha1(), NULL);
	FILE *fp = fopen((path + '/' + filename + ".fade").c_str(), "r");
	off_t fileSize;
	fread(&fileSize, sizeof(fileSize), 1, fp);
	off_t fileSizeB = (fileSize % AES_BLOCK_SIZE == 0) ? fileSize : ((fileSize / AES_BLOCK_SIZE + 1) * AES_BLOCK_SIZE);
	unsigned char buffer[FILE_BUFFER_LENGTH];
	unsigned int len;
	while ((len = fread(buffer, sizeof(unsigned char), FILE_BUFFER_LENGTH < fileSizeB ? FILE_BUFFER_LENGTH : fileSizeB, fp)) > 0)
	{
		HMAC_Update(&ctx, buffer, len);
		fileSizeB -= len;
	}
	unsigned char md[SHA_DIGEST_LENGTH];
	HMAC_Final(&ctx, md, &len);
	HMAC_CTX_cleanup(&ctx);
	unsigned char hmac[SHA_DIGEST_LENGTH];
	len = fread(hmac, sizeof(unsigned char), len, fp);
	fclose(fp);
	return (len == SHA_DIGEST_LENGTH && memcmp(md, hmac, SHA_DIGEST_LENGTH) == 0);
}

void Client::getFile(const string &filename, const string &ext) const
{
	for (vector<Storage *>::const_iterator it = storages.begin(); it != storages.end(); ++it)
	{
		if ((*it)->get(filename + ext))
		{
			break;
		}
	}
}

void Client::putFile(const string &filename, const string &ext) const
{
	for (vector<Storage *>::const_iterator it = storages.begin(); it != storages.end(); ++it)
	{
		(*it)->put(filename + ext);
	}
}

bool Client::upload(const string &filename, char *policyName)
{
	unsigned char key[AES_BLOCK_SIZE];
	RAND_bytes(key, AES_BLOCK_SIZE);
	FILE *fp = fopen((path + '/' + filename + ".metadata").c_str(), "w");
	char *token;
	const char *sep = ";";
	char *savePtr;
	for (char *str = policyName; ; str = NULL)
	{
		token = strtok_r(str, sep, &savePtr);
		if (!token)
		{
			break;
		}
		fprintf(fp, "%s\n", token);
		unsigned char k[AES_BLOCK_SIZE];
		memcpy(k, key, AES_BLOCK_SIZE);
		char *subToken;
		const char *subSep = ",";
		char *subSavePtr;
		for (char *subStr = token; ; subStr = NULL)
		{
			subToken = strtok_r(subStr, subSep, &subSavePtr);
			if (!subToken)
			{
				break;
			}
			PolicyForClient policy(subToken);
			const int rsaSize = policy.getRSASize();
			unsigned char *ki = reinterpret_cast<unsigned char *>(calloc(rsaSize, sizeof(unsigned char)));
			RAND_bytes(ki, AES_BLOCK_SIZE);
			AES_KEY aes;
			unsigned char iv[AES_BLOCK_SIZE];
			memset(iv, 0, sizeof(iv));
			if (AES_set_encrypt_key(ki, 128, &aes) < 0)
			{
				Panic::instance()->die("Unable to set encryption key in AES.\n");
			}
			unsigned char kki[AES_BLOCK_SIZE];
			AES_cbc_encrypt(k, kki, AES_BLOCK_SIZE, &aes, iv, AES_ENCRYPT);
			memcpy(k, kki, AES_BLOCK_SIZE);
			if (!policy.encrypt(ki))
			{
				fclose(fp);
				return false;
			}
			fwrite(ki, sizeof(unsigned char), rsaSize, fp);
			free(ki);
		}
		fwrite(k, sizeof(unsigned char), AES_BLOCK_SIZE, fp);
	}
	fclose(fp);
#ifdef FADE_EVAL
	ftime(&encStartTime);
#endif /* FADE_EVAL */
	encrypt(filename, key);
#ifdef FADE_EVAL
	ftime(&encEndTime);
	fileStartTime = encEndTime;
#endif /* FADE_EVAL */
	putFile(filename, ".fade");
#ifdef FADE_EVAL
	ftime(&fileEndTime);
	metaStartTime = fileEndTime;
#endif /* FADE_EVAL */
	putFile(filename, ".metadata");
#ifdef FADE_EVAL
	ftime(&metaEndTime);
#endif /* FADE_EVAL */
	return true;
}

bool Client::download(const string &filename)
{
#ifdef FADE_EVAL
	ftime(&fileStartTime);
#endif /* FADE_EVAL */
	getFile(filename, ".fade");
#ifdef FADE_EVAL
	ftime(&fileEndTime);
	hmacStartTime = fileEndTime;
#endif /* FADE_EVAL */
	if (!checkHMAC(filename))
	{
		Panic::instance()->warn("HMAC mismatch.\n");
		return false;
	}
#ifdef FADE_EVAL
	ftime(&hmacEndTime);
	metaStartTime = hmacEndTime;
#endif /* FADE_EVAL */
	getFile(filename, ".metadata");
#ifdef FADE_EVAL
	ftime(&metaEndTime);
#endif /* FADE_EVAL */
	FILE *fp = fopen((path + '/' + filename + ".metadata").c_str(), "r");
	char policyName[POLICY_NAME_LENGTH];
	while (fgets(policyName, POLICY_NAME_LENGTH, fp))
	{
		policyName[strlen(policyName) - 1] = '\0';
		Keychain *keychain = NULL;
		char *token;
		const char *sep = ",";
		char *savePtr;
		bool success = true;
		for (char *str = policyName; ; str = NULL)
		{
			token = strtok_r(str, sep, &savePtr);
			if (!token)
			{
				break;
			}
			PolicyForClient policy(token);
			const int rsaSize = policy.getRSASize();
			if (success)
			{
				unsigned char *key = reinterpret_cast<unsigned char *>(calloc(rsaSize, sizeof(unsigned char)));
				fread(key, sizeof(unsigned char), rsaSize, fp);
				if (!policy.decrypt(key))
				{
					success = false;
				}
				Keychain *kc = new Keychain;
				kc->key = key;
				kc->next = keychain;
				keychain = kc;
			}
			else
			{
				fseek(fp, rsaSize, SEEK_CUR);
			}
		}
		if (success)
		{
			unsigned char kki[AES_BLOCK_SIZE];
			fread(kki, sizeof(unsigned char), AES_BLOCK_SIZE, fp);
			while (keychain)
			{
				AES_KEY aes;
				unsigned char iv[AES_BLOCK_SIZE];
				memset(iv, 0, sizeof(iv));
				if (AES_set_decrypt_key(keychain->key, 128, &aes) < 0)
				{
					Panic::instance()->die("Unable to set decryption key in AES.\n");
				}
				unsigned char k[AES_BLOCK_SIZE];
				AES_cbc_encrypt(kki, k, AES_BLOCK_SIZE, &aes, iv, AES_DECRYPT);
				memcpy(kki, k, AES_BLOCK_SIZE);
				Keychain *kc = keychain;
				free(keychain->key);
				keychain = keychain->next;
				delete kc;
			}
			fclose(fp);
#ifdef FADE_EVAL
			ftime(&encStartTime);
#endif /* FADE_EVAL */
			decrypt(filename, kki);
#ifdef FADE_EVAL
			ftime(&encEndTime);
#endif /* FADE_EVAL */
			return true;
		}
		else
		{
			fseek(fp, AES_BLOCK_SIZE, SEEK_CUR);
			while (keychain)
			{
				Keychain *kc = keychain;
				free(keychain->key);
				keychain = keychain->next;
				delete kc;
			}
		}
	}
	fclose(fp);
	return false;
}

bool Client::renew(const string &filename, const char *newPolicyName)
{
#ifdef FADE_EVAL
	ftime(&metaStartTime);
#endif /* FADE_EVAL */
	getFile(filename, ".metadata");
#ifdef FADE_EVAL
	ftime(&metaEndTime);
#endif /* FADE_EVAL */
	FILE *fp = fopen((path + '/' + filename + ".metadata").c_str(), "r");
	char oldPolicyName[POLICY_NAME_LENGTH];
	fgets(oldPolicyName, POLICY_NAME_LENGTH, fp);
	oldPolicyName[strlen(oldPolicyName) - 1] = '\0';
	PolicyForClient oldPolicy(oldPolicyName);
	int rsaSize = oldPolicy.getRSASize();
	unsigned char *key = reinterpret_cast<unsigned char *>(calloc(rsaSize, sizeof(unsigned char)));
	fread(key, sizeof(unsigned char), rsaSize, fp);
	unsigned char k[AES_BLOCK_SIZE];
	fread(k, sizeof(unsigned char), AES_BLOCK_SIZE, fp);
	fclose(fp);
	if (!oldPolicy.decrypt(key))
	{
		return false;
	}
	PolicyForClient newPolicy(newPolicyName);
	rsaSize = newPolicy.getRSASize();
	if (!newPolicy.encrypt(key))
	{
		return false;
	}
	fp = fopen((path + '/' + filename + ".metadata").c_str(), "w");
	fprintf(fp, "%s\n", newPolicyName);
	fwrite(key, sizeof(unsigned char), rsaSize, fp);
	fwrite(k, sizeof(unsigned char), AES_BLOCK_SIZE, fp);
	fclose(fp);
#ifdef FADE_EVAL
	ftime(&metaStartTime2);
#endif /* FADE_EVAL */
	putFile(filename, ".metadata");
#ifdef FADE_EVAL
	ftime(&metaEndTime2);
#endif /* FADE_EVAL */
	return true;
}

bool Client::generateSecret()
{
	FILE *fp = fopen(secretFile.c_str(), "w");
	if (!fp)
	{
		return false;
	}
	RAND_bytes(secret, SECRET_LENGTH);
	fwrite(secret, sizeof(unsigned char), SECRET_LENGTH, fp);
	fclose(fp);
	return true;
}
