/*
 * FADE - Secure Overlay Cloud Storage with File Assured Deletion
 * Copyright (C) 2010  Yang Tang
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file Client.h
 * @author Yang Tang
 * @brief FADE client interface
 */

#ifndef FADE_CLIENT_H
#define FADE_CLIENT_H

#include <string>
#include <vector>
#include "LocalStorage.h"
#include "S3Storage.h"
#include "ClientConfig.h"

#ifdef FADE_EVAL
#include <sys/timeb.h>
#endif /* FADE_EVAL */

using namespace std;

static const int SECRET_LENGTH = 64;
static const int COMMAND_BUFFER_LENGTH = 256;

/**
 * @brief FADE client
 */
class Client
{
	int sockfd;
	string path;
	string ephemerizerAddress;
	vector<Storage *> storages;
	string secretFile;
	unsigned char secret[SECRET_LENGTH];
	Client() {}
	~Client()
	{
		for (vector<Storage *>::const_iterator it = storages.begin(); it != storages.end(); ++it)
		{
			delete *it;
		}
	}
	void closeSocket() const;
	void encrypt(const string &filename, const unsigned char *key) const;
	void decrypt(const string &filename, const unsigned char *key) const;
	bool checkHMAC(const string &filename) const;
	void getFile(const string &filename, const string& ext) const;
	void putFile(const string &filename, const string& ext) const;
public:
	/**
	 * Singleton instance
	 * @return The singleton instance of FADE client
	 */
	static Client *instance();

	/**
	 * Get the path of the local data cache
	 * @return The path of the local data cache
	 */
	string getPath() const
	{
		return path;
	}

	/**
	 * Set the path of the local data cache
	 * @param path The path of the local data cache
	 */
	void setPath(const string &path)
	{
		this->path = path;
	}

	/**
	 * Set the IP address of the key manager
	 * @param address The IP address of the key manager
	 */
	void setEphemerizerAddress(const string &address)
	{
		ephemerizerAddress = address;
	}

	/**
	 * Add a local storage
	 * @param dir The path of the local storage
	 */
	void addLocalStorage(const char *dir)
	{
		storages.push_back(new LocalStorage(dir));
	}

	/**
	 * Add an Amazon S3 storage
	 * @param bucket The bucket name of the Amazon S3 storage
	 */
	void addS3Storage(const char *bucket)
	{
		storages.push_back(new S3Storage(bucket));
	}

	/**
	 * Set the secret file
	 * @param secretFile The filename where the long-term private secret is stored
	 */
	void setSecretFile(const string &secretFile)
	{
		this->secretFile = secretFile;
	}

	/**
	 * Encrypt and upload a file with a policy
	 * @param filename The name of the file to be encrypted and uploaded
	 * @param policyName The name of the policy
	 * @return True if successful, otherwise false
	 */
	bool upload(const string &filename, char *policyName);

	/**
	 * Download and decrypt a file
	 * @param filename The name of the file to be downloaded and decrypted
	 * @return True if successful, otherwise false
	 */
	bool download(const string &filename);

	/**
	 * Renew a policy of a file
	 * @param filename The name of the file whose policy is to be renewed
	 * @param newPolicyName The name of the new policy
	 * @return True if successful, otherwise false
	 */
	bool renew(const string &filename, const char *newPolicyName);

	/**
	 * Generate a long-term private secret
	 * @return True if successful, otherwise false
	 */
	bool generateSecret();

	/**
	 * Read and parse the configuration file
	 * @param filename The name of the configuration file
	 */
	void readConfig(const char *filename)
	{
		ClientConfig::parse(filename);
	}

	/**
	 * Read the long-term private secret
	 */
	void readSecret();

	/**
	 * Interact with the key manager
	 * @param request The request data
	 * @param requestLength The length of the request
	 * @param response The response data
	 * @param responseLength The maximum size of the response buffer
	 * @return The actural length of the response
	 */
	unsigned int interact(const unsigned char *request, unsigned int requestLength, unsigned char *response, unsigned int responseLength);
#ifdef FADE_EVAL
	timeb encStartTime, encEndTime;
	timeb hmacStartTime, hmacEndTime;
	timeb allStartTime, allEndTime;
	timeb fileStartTime, fileEndTime;
	timeb metaStartTime, metaEndTime;
	timeb metaStartTime2, metaEndTime2;
#endif /* FADE_EVAL */
};

#endif /* FADE_CLIENT_H */
