/*
 * FADE - Secure Overlay Cloud Storage with File Assured Deletion
 * Copyright (C) 2010  Yang Tang
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file PolicyForClient.cc
 * @author Yang Tang
 * @brief Policy implementation used for FADE client
 */

#include <cstdio>
#include <cstring>
#include <openssl/pem.h>
#include <openssl/aes.h>
#include <openssl/md5.h>
#include "PolicyForClient.h"
#include "Client.h"
#include "../common/Panic.h"
#include "../common/Common.h"

PolicyForClient::PolicyForClient(const string &policyName)
	: policyName(policyName)
{
	ctx = BN_CTX_new();
	rsa = RSA_new();
}

bool PolicyForClient::hasPublicKey() const
{
	return (access((Client::instance()->getPath() + '/' + policyName + ".pubkey").c_str(), F_OK) == 0);
}

bool PolicyForClient::getPublicKey(bool allowNewPolicy) const
{
	const int rsaSize = getRSASize();
	if (!hasPublicKey())
	{
		unsigned char request[REQUEST_BUFFER_LENGTH];
		const int requestLength = snprintf(reinterpret_cast<char *>(request), REQUEST_BUFFER_LENGTH - rsaSize, allowNewPolicy ? "ENC %s" : "GET %s", policyName.c_str());
		unsigned char response[RESPONSE_BUFFER_LENGTH];
		const int n_e_size = Client::instance()->interact(request, requestLength, response, RESPONSE_BUFFER_LENGTH) - 1;
		if (!*response)
		{
			return false;
		}
		rsa->n = BN_bin2bn(response + 1, rsaSize, NULL);
		rsa->e = BN_bin2bn(response + 1 + rsaSize, n_e_size - rsaSize, NULL);
		FILE *fp = fopen((Client::instance()->getPath() + '/' + policyName + ".pubkey").c_str(), "w");
		fwrite(response + 1, sizeof(unsigned char), n_e_size, fp);
		fclose(fp);
	}
	else
	{
		unsigned char *buffer = reinterpret_cast<unsigned char *>(calloc(rsaSize, sizeof(unsigned char)));
		FILE *fp = fopen((Client::instance()->getPath() + '/' + policyName + ".pubkey").c_str(), "r");
		int readSize = fread(buffer, sizeof(unsigned char), rsaSize, fp);
		rsa->n = BN_bin2bn(buffer, readSize, NULL);
		readSize = fread(buffer, sizeof(unsigned char), rsaSize, fp);
		fclose(fp);
		rsa->e = BN_bin2bn(buffer, readSize, NULL);
		free(buffer);
	}
	return true;
}

bool PolicyForClient::encrypt(unsigned char *key) const
{
	if (!getPublicKey(true))
	{
		return false;
	}
	const int rsaSize = getRSASize();
	BIGNUM *k = BN_bin2bn(key, AES_BLOCK_SIZE, NULL);
	BIGNUM *ke = BN_new();
	BN_mod_exp(ke, k, rsa->e, rsa->n, ctx);
	BN_free(k);
	const int ke_len = BN_bn2bin(ke, key);
	if (ke_len < rsaSize) {
		memmove(key + rsaSize - ke_len, key, ke_len);
		memset(key, 0, rsaSize - ke_len);
	}
	BN_free(ke);
	return true;
}

bool PolicyForClient::decrypt(unsigned char *key) const
{
	if (!getPublicKey(false))
	{
		return false;
	}
	const int rsaSize = getRSASize();
	BIGNUM *ke = BN_bin2bn(key, rsaSize, NULL);
	BIGNUM *r = BN_new();
	BN_rand(r, 512, 1, true);
	BIGNUM *re = BN_new();
	BN_mod_exp(re, r, rsa->e, rsa->n, ctx);
	BIGNUM *kere = BN_new();
	BN_mod_mul(kere, ke, re, rsa->n, ctx);
	BN_free(ke);
	BN_free(re);
	unsigned char request[REQUEST_BUFFER_LENGTH];
	const int len = snprintf(reinterpret_cast<char *>(request), REQUEST_BUFFER_LENGTH - rsaSize, "DEC %s ", policyName.c_str());
	const int kere_size = BN_bn2bin(kere, request + len);
	BN_free(kere);
	unsigned char response[RESPONSE_BUFFER_LENGTH];
	const int kr_size = Client::instance()->interact(request, len + kere_size, response, RESPONSE_BUFFER_LENGTH) - 1;
	if (!*response)
	{
		return false;
	}
	BIGNUM *kr = BN_bin2bn(response + 1, kr_size, NULL);
	BIGNUM *k = BN_new();
	BN_div(k, NULL, kr, r, ctx);
	BN_free(r);
	BN_free(kr);
	const int k_len = BN_bn2bin(k, key);
	if (k_len < AES_BLOCK_SIZE){
		memmove(key + AES_BLOCK_SIZE - k_len, key, k_len);
		memset(key, 0, AES_BLOCK_SIZE - k_len);
	}
	BN_free(k);
	return true;
}
