/*
 * FADE - Secure Overlay Cloud Storage with File Assured Deletion
 * Copyright (C) 2010  Yang Tang
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file PolicyForClient.h
 * @author Yang Tang
 * @brief Policy interface used for FADE client
 */

#ifndef FADE_POLICY_FOR_CLIENT_H
#define FADE_POLICY_FOR_CLIENT_H

#include <string>
#include <openssl/rsa.h>

using namespace std;

/**
 * @brief Policy used for FADE client
 */
class PolicyForClient
{
	string policyName;
	RSA *rsa;
	BN_CTX *ctx;
	bool hasPublicKey() const;
	bool getPublicKey(bool allowNewPolicy) const;
public:
	/**
	 * @param policyName The name of the policy
	 */
	PolicyForClient(const string &policyName);

	~PolicyForClient()
	{
		RSA_free(rsa);
		BN_CTX_free(ctx);
	}

	/**
	 * Get the name of the policy
	 * @return The name of the policy
	 */
	string getPolicyName() const
	{
		return policyName;
	}

	/**
	 * Set the name of the policy
	 * @param policyName The name of the policy
	 */
	void setPolicyName(const string &policyName)
	{
		this->policyName = policyName;
	}

	/**
	 * Get RSA size used in OpenSSL
	 * @return RSA size
	 */
	int getRSASize() const
	{
		return 128;
	}

	/**
	 * Encrypt a key
	 * @param key The key to be encrypted
	 * @retval key The encrypted key
	 * @return True if successful, otherwise false
	 */
	bool encrypt(unsigned char *key) const;

	/**
	 * Decrypt a key
	 * @param key The key to be decrypted
	 * @retval key The decrypted key
	 * @return True if successful, otherwise false
	 */
	bool decrypt(unsigned char *key) const;
};

#endif /* FADE_POLICY_FOR_CLIENT_H */
