/*
 * FADE - Secure Overlay Cloud Storage with File Assured Deletion
 * Copyright (C) 2010  Yang Tang
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file S3Storage.cc
 * @author Yang Tang
 * @brief Amazon S3 storage implementation
 */

#include <cstdlib>
#include <fstream>
#include "S3Storage.h"
#include "Client.h"
#include "../common/Panic.h"

using namespace std;

static const int BUFFER_LENGTH = 4096;

S3Storage::S3Storage(const string &bucket)
	: bucket(bucket)
{
	char *accessKeyId = getenv("AWS_ACCESS_KEY_ID");
	char *secretAccessKey = getenv("AWS_SECRET_ACCESS_KEY");
	if (!accessKeyId || !secretAccessKey)
	{
		Panic::instance()->die("Please provide AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY.\n");
	}
	s3Rest = AWSConnectionFactory::getInstance()->createS3Connection(accessKeyId, secretAccessKey);
}

bool S3Storage::put(const string &filename) const
{
	try
	{
		ifstream inStream((Client::instance()->getPath() + '/' + filename).c_str());
		if (!inStream)
		{
			Panic::instance()->warn("Cannot open file for reading.\n");
			return false;
		}
		s3Rest->put(bucket, filename, inStream, "binary/octet-stream");
		inStream.close();
	}
	catch (PutException &e)
	{
		Panic::instance()->warn("Error putting file onto S3: %s\n", e.what());
		return false;
	}
	return true;
}

bool S3Storage::get(const string &filename) const
{
	try
	{
		GetResponsePtr get = s3Rest->get(bucket, filename);
		char buffer[BUFFER_LENGTH];
		istream &inStream = get->getInputStream();
		ofstream outStream((Client::instance()->getPath() + '/' + filename).c_str());
		if (!outStream)
		{
			Panic::instance()->warn("Cannot open file for writing.\n");
			return false;
		}
		while (inStream.good())
		{
			inStream.read(buffer, BUFFER_LENGTH);
			outStream.write(buffer, inStream.gcount());
		}
		outStream.close();
	}
	catch (GetException &e)
	{
		Panic::instance()->warn("Error getting file from S3: %s\n", e.what());
		return false;
	}
	return true;
}
