/*
 * FADE - Secure Overlay Cloud Storage with File Assured Deletion
 * Copyright (C) 2010  Yang Tang
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file Ephemerizer.cc
 * @author Yang Tang
 * @brief FADE key manager implementation
 */

#include <cerrno>
#include <cstring>
#include <sys/socket.h>
#include <netinet/in.h>
#include <openssl/bn.h>
#include "Ephemerizer.h"
#include "../common/Panic.h"
#include "../common/Common.h"

Ephemerizer *Ephemerizer::instance()
{
	static Ephemerizer _instance;
	return &_instance;
}

void Ephemerizer::initSocket()
{
	sockaddr_in addr;
	addr.sin_port = htons(DEFAULT_PORT);
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_family = AF_INET;
	if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		Panic::instance()->die(strerror(errno));
	}
	const int reuseaddr = 1;
	if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &reuseaddr, sizeof(int)))
	{
		Panic::instance()->warn(strerror(errno));
	}
	if (bind(sockfd, reinterpret_cast<sockaddr *>(&addr), sizeof(addr)))
	{
		closeSocket();
		Panic::instance()->die(strerror(errno));
	}
}

void Ephemerizer::closeSocket() const
{
	if (close(sockfd))
	{
		Panic::instance()->die(strerror(errno));
	}
}

void Ephemerizer::run() const
{
	unsigned char request[REQUEST_BUFFER_LENGTH];
	ssize_t len;
	socklen_t alen;
	char abuf[ADDRESS_LENGTH];
	for (;;)
	{
		alen = ADDRESS_LENGTH;
		if ((len = recvfrom(sockfd, request, sizeof(request), 0, reinterpret_cast<sockaddr *>(abuf), &alen)) > 0)
		{
			unsigned char *response;
			int keyLength = 0;
			if (memcmp(request, "ENC ", 4) == 0)
			{
				const string policyName(reinterpret_cast<char *>(request + 4), len - 4);
				PolicyForEphemerizer policy(policyName);
				BIGNUM *n = policy.getN();
				BIGNUM *e = policy.getE();
				response = reinterpret_cast<unsigned char *>(calloc(BN_num_bytes(n) + BN_num_bytes(e) + 1, sizeof(unsigned char)));
				*response = 1;
				keyLength = BN_bn2bin(n, response + 1);
				keyLength += BN_bn2bin(e, response + 1 + keyLength);
			}
			else if (memcmp(request, "GET ", 4) == 0)
			{
				const string policyName(reinterpret_cast<char *>(request + 4), len - 4);
				if (PolicyForEphemerizer::exist(policyName))
				{
					PolicyForEphemerizer policy(policyName);
					BIGNUM *n = policy.getN();
					BIGNUM *e = policy.getE();
					response = reinterpret_cast<unsigned char *>(calloc(BN_num_bytes(n) + BN_num_bytes(e) + 1, sizeof(unsigned char)));
					*response = 1;
					keyLength = BN_bn2bin(n, response + 1);
					keyLength += BN_bn2bin(e, response + 1 + keyLength);
				}
				else
				{
					response = reinterpret_cast<unsigned char *>(calloc(1, sizeof(unsigned char)));
				}
			}
			else if (memcmp(request, "DEC ", 4) == 0)
			{
                                unsigned char *key = reinterpret_cast<unsigned char *>(memchr(request + 4, ' ', len - 4)) + 1;
				const string policyName(reinterpret_cast<char *>(request + 4), key - request - 5);
				if (PolicyForEphemerizer::exist(policyName))
				{
					PolicyForEphemerizer policy(policyName);
					keyLength = policy.decrypt(len - (key - request), key, key);
					response = reinterpret_cast<unsigned char *>(calloc(keyLength + 1, sizeof(unsigned char)));
					*response = 1;
					memcpy(response + 1, key, keyLength);
				}
				else
				{
					response = reinterpret_cast<unsigned char *>(calloc(1, sizeof(unsigned char)));
				}
			}
			else
			{
				continue;
			}
			if (sendto(sockfd, response, keyLength + 1, 0, reinterpret_cast<sockaddr *>(abuf), alen) < 0)
			{
				Panic::instance()->warn(strerror(errno));
			}
			free(response);
		}
	}
}
