/*
 * FADE - Secure Overlay Cloud Storage with File Assured Deletion
 * Copyright (C) 2010  Yang Tang
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file PolicyForEphemerizer.h
 * @author Yang Tang
 * @brief Policy interface for FADE key manager
 */

#ifndef FADE_POLICY_FOR_EPHEMERIZER_H
#define FADE_POLICY_FOR_EPHEMERIZER_H

#include <openssl/rsa.h>
#include <unistd.h>
#include <string>

using namespace std;

/**
 * @brief Policy for FADE key manager
 */
class PolicyForEphemerizer
{
	string policyName;
	RSA *rsa;
	BN_CTX *ctx;
	void generateKey();
	void readPEM();
	void writePEM() const;
public:
	/**
	 * @param policyName The name of the policy
	 */
	PolicyForEphemerizer(const string &policyName);

	~PolicyForEphemerizer()
	{
		RSA_free(rsa);
		BN_CTX_free(ctx);
	}

	/**
	 * Get the name of the policy
	 * @return The name of the policy
	 */
	string getPolicyName() const
	{
		return policyName;
	}

	/**
	 * Get the RSA size used in OpenSSL
	 * @return RSA size
	 */
	int getRSASize() const
	{
		return RSA_size(rsa);
	}

	/**
	 * Get n used in RSA
	 * @return n
	 */
	BIGNUM *getN() const
	{
		return rsa->n;
	}

	/**
	 * Get e used in RSA
	 * @return e
	 */
	BIGNUM *getE() const
	{
		return rsa->e;
	}

	/**
	 * Decrypt a message
	 * @param flen The length of the message to be decrypted
	 * @param from The message to be decrypted
	 * @retval to The decrypted message
	 * @return The length of the decrypted message
	 */
	int decrypt(int flen, const unsigned char *from, unsigned char *to) const;

	/**
	 * Check if a policy already exists or not
	 * @param policyName The name of the policy
	 * @return True if the policy already exists, otherwise false
	 */
	static bool exist(const string &policyName)
	{
		return (access(("keys/" + policyName + ".pem").c_str(), F_OK) == 0);
	}
};

#endif /* FADE_POLICY_FOR_EPHEMERIZER_H */
